@@ -3,12 +3,14 @@
 import logging
 import math
 from abc import ABC
-from typing import Any, Dict, Iterator, List, Optional, Type, Union, cast
+from typing import Any, Dict, Iterator, List, Optional, Type, Union, cast, overload
 
 import dask.array as da
 import numpy as np
 from dask import delayed
 
+from .axes import Axes
+from .format import format_from_version
 from .io import ZarrLocation
 from .types import JSONDict
 
@@ -55,10 +57,24 @@ class Node:
             self.specs.append(PlateLabels(self))
         elif Plate.matches(zarr):
             self.specs.append(Plate(self))
-            self.add(zarr, plate_labels=True)
+            # self.add(zarr, plate_labels=True)
         if Well.matches(zarr):
             self.specs.append(Well(self))
 
+    @overload
+    def first(self, spectype: Type["Well"]) -> Optional["Well"]:
+        ...
+
+    @overload
+    def first(self, spectype: Type["Plate"]) -> Optional["Plate"]:
+        ...
+
+    def first(self, spectype: Type["Spec"]) -> Optional["Spec"]:
+        for spec in self.specs:
+            if isinstance(spec, spectype):
+                return spec
+        return None
+
     @property
     def visible(self) -> bool:
         """True if this node should be displayed by default.
@@ -261,20 +277,22 @@ class Multiscales(Spec):
     def __init__(self, node: Node) -> None:
         super().__init__(node)
 
-        axes_values = {"t", "c", "z", "y", "x"}
         try:
             multiscales = self.lookup("multiscales", [])
             version = multiscales[0].get(
                 "version", "0.1"
             )  # should this be matched with Format.version?
             datasets = multiscales[0]["datasets"]
-            # axes field was introduced in 0.3, before all data was 5d
-            axes = tuple(multiscales[0].get("axes", ["t", "c", "z", "y", "x"]))
-            if len(set(axes) - axes_values) > 0:
-                raise RuntimeError(f"Invalid axes names: {set(axes) - axes_values}")
-            node.metadata["axes"] = axes
-            datasets = [d["path"] for d in datasets]
-            self.datasets: List[str] = datasets
+            axes = multiscales[0].get("axes")
+            fmt = format_from_version(version)
+            # Raises ValueError if not valid
+            axes_obj = Axes(axes, fmt)
+            node.metadata["axes"] = axes_obj.to_list()
+            paths = [d["path"] for d in datasets]
+            self.datasets: List[str] = paths
+            transformations = [d.get("coordinateTransformations") for d in datasets]
+            if any(trans is not None for trans in transformations):
+                node.metadata["coordinateTransformations"] = transformations
             LOGGER.info("datasets %s", datasets)
         except Exception as e:
             LOGGER.error(f"failed to parse multiscale metadata: {e}")
@@ -287,7 +305,12 @@ class Multiscales(Spec):
                 for c in data.chunks
             ]
             LOGGER.info("resolution: %s", resolution)
-            LOGGER.info(" - shape %s = %s", axes, data.shape)
+            axes_names = None
+            if axes is not None:
+                axes_names = tuple(
+                    axis if isinstance(axis, str) else axis["name"] for axis in axes
+                )
+            LOGGER.info(" - shape %s = %s", axes_names, data.shape)
             LOGGER.info(" - chunks =  %s", chunk_sizes)
             LOGGER.info(" - dtype = %s", data.dtype)
             node.data.append(data)
@@ -389,9 +412,13 @@ class Well(Spec):
         # Use first Field for rendering settings, shape etc.
         image_zarr = self.zarr.create(image_paths[0])
         image_node = Node(image_zarr, node)
+        x_index = len(image_node.metadata["axes"]) - 1
+        y_index = len(image_node.metadata["axes"]) - 2
         level = 0  # load full resolution image
-        numpy_type = image_node.data[level].dtype
-        img_shape = image_node.data[level].shape
+        self.numpy_type = image_node.data[level].dtype
+        self.img_metadata = image_node.metadata
+        self.img_shape = image_node.data[level].shape
+        self.img_pyramid_shapes = [d.shape for d in image_node.data]
 
         # stitch full-resolution images into a grid
         def get_field(tile_name: str) -> np.ndarray:
@@ -404,7 +431,7 @@ class Well(Spec):
                 data = self.zarr.load(path)
             except ValueError:
                 LOGGER.error(f"Failed to load {path}")
-                data = np.zeros(img_shape, dtype=numpy_type)
+                data = np.zeros(self.img_shape, dtype=self.numpy_type)
             return data
 
         lazy_reader = delayed(get_field)
@@ -418,11 +445,13 @@ class Well(Spec):
                     tile_name = f"{row},{col}"
                     LOGGER.debug(f"creating lazy_reader. row:{row} col:{col}")
                     lazy_tile = da.from_delayed(
-                        lazy_reader(tile_name), shape=img_shape, dtype=numpy_type
+                        lazy_reader(tile_name),
+                        shape=self.img_shape,
+                        dtype=self.numpy_type,
                     )
                     lazy_row.append(lazy_tile)
-                lazy_rows.append(da.concatenate(lazy_row, axis=4))
-            return da.concatenate(lazy_rows, axis=3)
+                lazy_rows.append(da.concatenate(lazy_row, axis=x_index))
+            return da.concatenate(lazy_rows, axis=y_index)
 
         node.data = [get_lazy_well()]
         node.metadata = image_node.metadata
@@ -443,10 +472,9 @@ class Plate(Spec):
         stitched full-resolution images.
         """
         self.plate_data = self.lookup("plate", {})
-        LOGGER.info("plate_data", self.plate_data)
+        LOGGER.info("plate_data: %s", self.plate_data)
         self.rows = self.plate_data.get("rows")
         self.columns = self.plate_data.get("columns")
-        self.acquisitions = self.plate_data.get("acquisitions")
         self.first_field = "0"
         self.row_names = [row["name"] for row in self.rows]
         self.col_names = [col["name"] for col in self.columns]
@@ -454,30 +482,22 @@ class Plate(Spec):
         self.well_paths = [well["path"] for well in self.plate_data.get("wells")]
         self.well_paths.sort()
 
-        self.run = ""
-        # TEMP - support acquisition path in plate/acq/row/col hierarchy
-        # remove when we don't want to support dev versions of ome-zarr plate data
-        if len(self.acquisitions) > 0:
-            self.run = self.acquisitions[0].get("path", "")
-            if len(self.run) > 0 and not self.run.endswith("/"):
-                self.run = self.run + "/"
-
         self.row_count = len(self.rows)
         self.column_count = len(self.columns)
-        # Get the first image...
-        path = f"{self.well_paths[0]}/{self.first_field}"
-        image_zarr = self.zarr.create(path)
-        image_node = Node(image_zarr, node)
 
-        self.numpy_type = self.get_numpy_type(image_node)
-        img_shape = image_node.data[0].shape
+        # Get the first well...
+        well_zarr = self.zarr.create(self.well_paths[0])
+        well_node = Node(well_zarr, node)
+        well_spec: Optional[Well] = well_node.first(Well)
+        if well_spec is None:
+            raise Exception("could not find first well")
+        self.numpy_type = well_spec.numpy_type
 
-        img_pyramid_shapes = [d.shape for d in image_node.data]
+        LOGGER.debug(f"img_pyramid_shapes: {well_spec.img_pyramid_shapes}")
 
-        LOGGER.debug("img_pyramid_shapes", img_pyramid_shapes)
-
-        size_y = img_shape[3]
-        size_x = img_shape[4]
+        self.axes = well_spec.img_metadata["axes"]
+        size_y = well_spec.img_shape[len(self.axes) - 2]
+        size_x = well_spec.img_shape[len(self.axes) - 1]
 
         # FIXME - if only returning a single stiched plate (not a pyramid)
         # need to decide optimal size. E.g. longest side < 1500
@@ -486,7 +506,7 @@ class Plate(Spec):
         plate_height = self.row_count * size_y
         longest_side = max(plate_width, plate_height)
         target_level = 0
-        for level, shape in enumerate(img_pyramid_shapes):
+        for level, shape in enumerate(well_spec.img_pyramid_shapes):
             plate_width = self.column_count * shape[-1]
             plate_height = self.row_count * shape[-2]
             longest_side = max(plate_width, plate_height)
@@ -494,7 +514,7 @@ class Plate(Spec):
             if longest_side <= TARGET_SIZE:
                 break
 
-        LOGGER.debug("target_level", target_level)
+        LOGGER.debug(f"target_level: {target_level}")
 
         pyramid = []
 
@@ -502,14 +522,14 @@ class Plate(Spec):
         # for level in range(5):
         for level in [target_level]:
 
-            tile_shape = img_pyramid_shapes[level]
+            tile_shape = well_spec.img_pyramid_shapes[level]
             lazy_plate = self.get_stitched_grid(level, tile_shape)
             pyramid.append(lazy_plate)
 
         # Set the node.data to be pyramid view of the plate
         node.data = pyramid
         # Use the first image's metadata for viewing the whole Plate
-        node.metadata = image_node.metadata
+        node.metadata = well_spec.img_metadata
 
         # "metadata" dict gets added to each 'plate' layer in napari
         node.metadata.update({"metadata": {"plate": self.plate_data}})
@@ -519,21 +539,24 @@ class Plate(Spec):
 
     def get_tile_path(self, level: int, row: int, col: int) -> str:
         return (
-            f"{self.run}{self.row_names[row]}/"
+            f"{self.row_names[row]}/"
             f"{self.col_names[col]}/{self.first_field}/{level}"
         )
 
     def get_stitched_grid(self, level: int, tile_shape: tuple) -> da.core.Array:
+        LOGGER.debug(f"get_stitched_grid() level: {level}, tile_shape: {tile_shape}")
+
         def get_tile(tile_name: str) -> np.ndarray:
             """tile_name is 'level,z,c,t,row,col'"""
             row, col = (int(n) for n in tile_name.split(","))
             path = self.get_tile_path(level, row, col)
-            LOGGER.debug(f"LOADING tile... {path}")
+            LOGGER.debug(f"LOADING tile... {path} with shape: {tile_shape}")
 
             try:
                 data = self.zarr.load(path)
-            except ValueError:
+            except ValueError as e:
                 LOGGER.error(f"Failed to load {path}")
+                LOGGER.debug(f"{e}")
                 data = np.zeros(tile_shape, dtype=self.numpy_type)
             return data
 
@@ -549,12 +572,12 @@ class Plate(Spec):
                     lazy_reader(tile_name), shape=tile_shape, dtype=self.numpy_type
                 )
                 lazy_row.append(lazy_tile)
-            lazy_rows.append(da.concatenate(lazy_row, axis=4))
-        return da.concatenate(lazy_rows, axis=3)
+            lazy_rows.append(da.concatenate(lazy_row, axis=len(self.axes) - 1))
+        return da.concatenate(lazy_rows, axis=len(self.axes) - 2)
 
 
 class PlateLabels(Plate):
-    def get_tile_path(self, level: int, row: int, col: int) -> str:
+    def get_tile_path(self, level: int, row: int, col: int) -> str:  # pragma: no cover
         """251.zarr/A/1/0/labels/0/3/"""
         path = (
             f"{self.row_names[row]}/{self.col_names[col]}/"
@@ -562,10 +585,16 @@ class PlateLabels(Plate):
         )
         return path
 
-    def get_pyramid_lazy(self, node: Node) -> None:
+    def get_pyramid_lazy(self, node: Node) -> None:  # pragma: no cover
         super().get_pyramid_lazy(node)
         # pyramid data may be multi-channel, but we only have 1 labels channel
-        node.data[0] = node.data[0][:, 0:1, :, :, :]
+        # TODO: when PlateLabels are re-enabled, update the logic to handle
+        # 0.4 axes (list of dictionaries)
+        if "c" in self.axes:
+            c_index = self.axes.index("c")
+            idx = [slice(None)] * len(self.axes)
+            idx[c_index] = slice(0, 1)
+            node.data[0] = node.data[0][tuple(idx)]
         # remove image metadata
         node.metadata = {}
 
@@ -585,7 +614,7 @@ class PlateLabels(Plate):
                         del properties[label_val]["label-value"]
         node.metadata["properties"] = properties
 
-    def get_numpy_type(self, image_node: Node) -> np.dtype:
+    def get_numpy_type(self, image_node: Node) -> np.dtype:  # pragma: no cover
         # FIXME - don't assume Well A1 is valid
         path = self.get_tile_path(0, 0, 0)
         label_zarr = self.zarr.load(path)
